import json
from abc import ABC
from collections.abc import Iterable
from copy import copy
from io import BytesIO
from json.decoder import JSONDecodeError
from typing import Any, Dict, ItemsView, KeysView, Optional, ValuesView

import yaml

from .http_multipart_message_parser import parse_multipart_message
from .http_query_string import parse_qs


class HttpMessage(ABC):
    pass


class SimpleHttpMessage(HttpMessage, str):
    def __new__(cls, value, *args, **kwargs):
        return super(SimpleHttpMessage, cls).__new__(cls, value)

    def __init__(self, value: str):
        # ignore value, as strings are immutable and this should be passed to __new__ constructor
        ...


class CompositeHttpMessage(HttpMessage):
    def __init__(self, data: Any) -> None:
        self.data = data

    def get(self, name: str, default: Optional[Any] = None) -> Any:
        if name in self:
            return self[name]

        return default

    def __getitem__(self, key) -> Any:
        return self.data[key]

    def __contains__(self, key):
        return key in self.data

    def __copy__(self):
        return type(self)(copy(self.data))

    def __iter__(self):
        return iter(self.data)

    def items(self) -> ItemsView:
        return self.data.items()

    def values(self) -> ValuesView:
        return self.data.values()

    def keys(self) -> KeysView:
        return self.data.keys()

    def __int__(self):
        return int(self.data)

    def __str__(self):
        return str(self.data)

    def __float__(self):
        return float(self.data)

    def __bool__(self):
        return bool(self.data)


class YamlHttpMessage(CompositeHttpMessage):
    @staticmethod
    def from_bytes(body: BytesIO, encoding: str = "utf8") -> "YamlHttpMessage":
        body.seek(0)
        decoded_input = body.read().decode(encoding)

        parsed_body: Dict[str, Any] = {}
        try:
            parsed_body = yaml.safe_load_all(decoded_input)  # type: ignore
        except JSONDecodeError:
            ...  # ignore

        return YamlHttpMessage(parsed_body)


class FormHttpMessage(CompositeHttpMessage):
    @staticmethod
    def from_bytes(body: BytesIO, encoding: str = "utf8") -> "FormHttpMessage":
        body.seek(0)
        decoded_input = body.read().decode(encoding)
        fields = parse_qs(decoded_input)
        return FormHttpMessage(fields)


class JsonHttpMessage(CompositeHttpMessage):
    @staticmethod
    def from_bytes(body: BytesIO, encoding: str = "utf8") -> "JsonHttpMessage":
        body.seek(0)
        decoded_input = body.read().decode(encoding)

        parsed_body: Dict[str, Any] = {}
        try:
            parsed_body = json.loads(decoded_input)
        except JSONDecodeError:
            ...  # ignore

        return JsonHttpMessage(parsed_body)


class MultipartHttpMessage(CompositeHttpMessage):
    @staticmethod
    def from_bytes(body: BytesIO, boundary: str, encoding: str = "utf8") -> "MultipartHttpMessage":
        body.seek(0)
        fields = parse_multipart_message(body.read(), boundary, encoding)

        return MultipartHttpMessage(fields)


__all__ = [
    "HttpMessage",
    "CompositeHttpMessage",
    "FormHttpMessage",
    "JsonHttpMessage",
    "MultipartHttpMessage",
    "SimpleHttpMessage",
    "YamlHttpMessage",
]



import inspect
from typing import Callable, Dict, Any, List, Optional, Union

from chocs.http_method import HttpMethod
from chocs.http_request import HttpRequest
from chocs.http_response import HttpResponse
from chocs.json_schema.json_schema import JsonReference, OpenApiSchema
from chocs.json_schema.schema_validator import build_validator_from_schema
from chocs.json_schema.errors import InvalidInputValidationError
from chocs.middleware import Middleware, MiddlewareHandler
from chocs.routing import Route
from chocs.http_message import JsonHttpMessage


class OpenApiMiddleware(Middleware):
    def __init__(self, openapi_filename: str, validate_body: bool = True, validate_query: bool = True):
        self.openapi = OpenApiSchema(openapi_filename)
        self.validate_body = validate_body
        self.validate_query = validate_query
        self.body_validators: Dict[str, Optional[Callable]] = {}
        self.query_validators: Dict[str, Optional[Callable]] = {}

    def handle(self, request: HttpRequest, next: MiddlewareHandler) -> HttpResponse:
        route = request.route
        if not route:
            return next(request)

        if self.validate_body:
            self._validate_request_body(request, route)

        if self.validate_query:
            self._validate_query_string(request, route)

        return next(request)

    def _validate_query_string(self, request: HttpRequest, route: Route) -> None:
        query_validator = self._get_query_validator(route.route, request.method)
        if not query_validator:
            return

        query_validator(dict(request.query_string) if request.query_string else {})

    def _validate_request_body(self, request: HttpRequest, route: Route) -> None:
        body_validator = self._get_body_validator(route.route, request.method)
        if not body_validator:
            return

        if not isinstance(request.parsed_body, JsonHttpMessage):
            raise InvalidInputValidationError(message="Request body is not valid application/json request.")

        parsed_body: Union[Dict, List]

        # we expect here only types provided by json parser, we should ignore everything what is not dict or list.
        if isinstance(request.parsed_body.data, dict) or isinstance(request.parsed_body.data, list):
            parsed_body = request.parsed_body.data
        else:
            raise InvalidInputValidationError(message="Request body could not be validated.")

        valid_body = body_validator(parsed_body)

        if "parsed_body" not in route.attributes:
            return

        if not inspect.isclass(route.attributes["parsed_body"]):
            return

        strict = route.attributes["strict"] if "strict" in route.attributes else True
        constructor = route.attributes["parsed_body"]
        request._parsed_body = None

        if not strict:

            def _get_non_strict_parsed_body() -> Any:

                instance = constructor.__new__(constructor)
                for prop_name, prop_value in valid_body.items():
                    setattr(instance, prop_name, prop_value)

                if hasattr(instance, "__post_init__"):
                    instance.__post_init__()

                return instance

            request._parsed_body_getter = _get_non_strict_parsed_body

            return

        def _get_strict_parsed_body() -> Any:
            return constructor(**valid_body)

        request._parsed_body_getter = _get_strict_parsed_body

    def _get_query_validator(self, route: str, method: HttpMethod) -> Optional[Callable]:
        if route not in self.query_validators:
            path = route.replace("/", "\\/")
            method_name = str(method).lower()
            try:
                query_parameters = [
                    item
                    for item in self.openapi.query(f"/paths/{path}/{method_name}/parameters")
                    if item["in"] == "query"
                ]
                query_schema = {
                    "type": "object",
                    "properties": {},
                    "required": [],
                }
                for param in query_parameters:
                    if isinstance(param, JsonReference):
                        param = param.data
                    param_name = param.get("name", "_")
                    query_schema["properties"][param_name] = param["schema"]  # type:ignore
                    if param["required"]:
                        query_schema["required"].append(param_name)  # type:ignore

                self.query_validators[route] = build_validator_from_schema(query_schema)
            except KeyError:
                self.query_validators[route] = None

        return self.query_validators[route]

    def _get_body_validator(self, route: str, method: HttpMethod) -> Optional[Callable]:
        if route not in self.body_validators:
            path = route.replace("/", "\\/")
            method_name = str(method).lower()
            try:
                body_schema = self.openapi.query(
                    f"/paths/{path}/{method_name}/requestBody/content/application\\/json/schema"
                )
                self.body_validators[route] = build_validator_from_schema(body_schema)
            except KeyError:
                self.body_validators[route] = None

        return self.body_validators[route]


__all__ = ["OpenApiMiddleware"]



from typing import Any, Callable, Iterable

from .errors import ValidationError


def validate_all_of(value: Any, validators: Iterable[Callable]) -> Any:
    for validate in validators:
        value = validate(value)
    return value


def validate_any_of(value: Any, validators: Iterable[Callable]) -> Any:
    for validate in validators:
        try:
            validate(value)
            return value
        except ValueError:
            continue

    raise ValidationError("Value could not be validated", code="any_error")


def validate_one_of(value: Any, validators: Iterable[Callable]) -> Any:
    valid_count = 0
    for validate in validators:
        try:
            validate(value)
            valid_count += 1
        except ValueError:
            continue

        if valid_count > 1:
            raise ValidationError("Value should only match one of validators", code="one_of_error")

    return value


def validate_not(value: Any, validator: Callable) -> Any:
    try:
        validator(value)
        raise ValidationError("Value should not match passed validator", code="not_error")
    except ValueError:
        return value


__all__ = ["validate_any_of", "validate_all_of", "validate_not", "validate_one_of"]



import re
from typing import Any

from .errors import FormatValidationError, MaximumLengthError, MinimumLengthError
from .string_format import StringFormat
from .type_validators import validate_string


def validate_minimum_length(value: str, expected_minimum: int) -> str:
    validate_string(value)
    if len(value) >= expected_minimum:
        return value

    raise MinimumLengthError(expected_minimum=expected_minimum)


def validate_maximum_length(value: str, expected_maximum: int) -> str:
    validate_string(value)
    if len(value) <= expected_maximum:
        return value
    raise MaximumLengthError(expected_maximum=expected_maximum)


def validate_string_pattern(value: str, pattern: str) -> str:
    if not re.search(pattern, value):
        raise FormatValidationError(expected_format=pattern)

    return value


def validate_string_format(value: str, format_name: str) -> Any:
    format_validator = StringFormat[format_name]

    return format_validator(value)


__all__ = [
    "validate_maximum_length",
    "validate_minimum_length",
    "validate_string_format",
    "validate_string_pattern",
]



from .array_validators import validate_items, validate_maximum_items, validate_minimum_items, validate_unique_items
from .combining_validators import validate_all_of, validate_any_of, validate_not, validate_one_of
from .number_validators import (
    validate_exclusive_maximum,
    validate_exclusive_minimum,
    validate_maximum,
    validate_minimum,
    validate_multiple_of,
)
from .object_validators import validate_object_properties
from .string_validators import (
    validate_maximum_length,
    validate_minimum_length,
    validate_string_format,
    validate_string_pattern,
)
from .type_validators import (
    validate_array,
    validate_boolean,
    validate_enum,
    validate_integer,
    validate_nullable,
    validate_number,
    validate_string,
)

__all__ = [
    "validate_array",
    "validate_boolean",
    "validate_enum",
    "validate_integer",
    "validate_nullable",
    "validate_number",
    "validate_string",
    "validate_items",
    "validate_maximum_items",
    "validate_minimum_items",
    "validate_unique_items",
    "validate_minimum_length",
    "validate_maximum_length",
    "validate_string_format",
    "validate_string_pattern",
    "validate_exclusive_maximum",
    "validate_exclusive_minimum",
    "validate_maximum",
    "validate_minimum",
    "validate_multiple_of",
    "validate_any_of",
    "validate_all_of",
    "validate_not",
    "validate_one_of",
    "validate_object_properties",
]



from .aws import AwsContext, AwsEvent, AwsServerlessFunction, create_http_request_from_aws_event
from .serverless import IS_AWS_ENVIRONMENT, ServerlessFunction
from .wrapper import create_serverless_function



from typing import Callable

from chocs.json_schema.format_validators import (
    validate_format_boolean,
    validate_format_bytes,
    validate_format_date,
    validate_format_datetime,
    validate_format_decimal,
    validate_format_email,
    validate_format_hostname,
    validate_format_ip_address,
    validate_format_ip_address_v4,
    validate_format_ip_address_v6,
    validate_format_pattern,
    validate_format_semver,
    validate_format_time,
    validate_format_time_duration,
    validate_format_uri,
    validate_format_url,
    validate_format_uuid,
)


class _StringFormat:
    BOOLEAN = "boolean"
    BYTE = "byte"
    DATE = "date"
    DATE_TIME = "date-time"
    DECIMAL = "decimal"
    EMAIL = "email"
    HOSTNAME = "hostname"
    IP_ADDRESS = "ip-address"
    IP_ADDRESS_V4 = "ip-address-v4"
    IP_ADDRESS_V6 = "ip-address-v6"
    PATTERN = "pattern"
    SEMVER = "semver"
    TIME = "time"
    TIME_DURATION = "time-duration"
    URI = "uri"
    URL = "url"
    UUID = "uuid"

    def __init__(self):
        self._formats = {
            self.BOOLEAN: validate_format_boolean,
            self.BYTE: validate_format_bytes,
            self.DATE: validate_format_date,
            self.DATE_TIME: validate_format_datetime,
            self.DECIMAL: validate_format_decimal,
            self.EMAIL: validate_format_email,
            self.HOSTNAME: validate_format_hostname,
            self.IP_ADDRESS: validate_format_ip_address,
            self.IP_ADDRESS_V4: validate_format_ip_address_v4,
            self.IP_ADDRESS_V6: validate_format_ip_address_v6,
            self.PATTERN: validate_format_pattern,
            self.SEMVER: validate_format_semver,
            self.TIME: validate_format_time,
            self.TIME_DURATION: validate_format_time_duration,
            self.URI: validate_format_uri,
            self.URL: validate_format_url,
            self.UUID: validate_format_uuid,
        }

    def __getitem__(self, format_name: str) -> Callable:
        if format_name in self._formats:
            return self._formats[format_name]
        raise KeyError(f"Unsupported format {format_name}")

    def __setitem__(self, format_name: str, value: Callable) -> None:
        self._formats[format_name] = value

    def __contains__(self, format_name: str) -> bool:
        return format_name in self._formats


StringFormat = _StringFormat()


__all__ = ["StringFormat"]



from typing import Callable, List, Set, Union

from .errors import AdditionalItemsError, MaximumLengthError, MinimumLengthError, UniqueItemsValidationError
from .type_validators import validate_array


def validate_unique_items(value: list) -> list:
    validate_array(value)

    unique_items: Set = set()
    for item in value:
        for unique_item in unique_items:
            if item is unique_item:
                raise UniqueItemsValidationError()

        unique_items.add(item)

    return value


def validate_minimum_items(value: list, expected_minimum: int) -> list:
    validate_array(value)

    if len(value) >= expected_minimum:
        return value

    raise MinimumLengthError(expected_minimum=expected_minimum)


def validate_maximum_items(value: list, expected_maximum: int) -> list:
    validate_array(value)

    if len(value) <= expected_maximum:
        return value

    raise MaximumLengthError(expected_maximum=expected_maximum)


def validate_items(value: list, item_validator: Callable) -> list:
    validate_array(value)

    return [item_validator(item) for item in value]


def validate_tuple(value: list, item_validator: List[Callable], additional_items: Union[bool, Callable]) -> list:
    list_length = len(value)
    validators_length = len(item_validator)
    additional_items_validator = additional_items if callable(additional_items) else lambda x: x

    if list_length > validators_length and additional_items is False:
        raise AdditionalItemsError()

    for i in range(0, list_length):
        if i < validators_length:
            value[i] = item_validator[i](value[i])
            continue

        value[i] = additional_items_validator(value[i])  # type: ignore

    return value


__all__ = [
    "validate_items",
    "validate_maximum_items",
    "validate_minimum_items",
    "validate_unique_items",
    "validate_tuple",
]



from enum import Enum


class HttpMethod(Enum):
    GET = "GET"
    HEAD = "HEAD"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    CONNECT = "CONNECT"
    OPTIONS = "OPTIONS"
    TRACE = "TRACE"
    PATCH = "PATCH"

    def __str__(self) -> str:
        return str(self.value)


__all__ = ["HttpMethod"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chocs', 'chocs.json_schema', 'chocs.middleware', 'chocs.serverless']

package_data = \
{'': ['*']}

install_requires = \
['pyyaml>=5.3.1,<6.0.0', 'typing_extensions>=3.7.4,<4.0.0']

setup_kwargs = {
    'name': 'chocs',
    'version': '0.9.6',
    'description': 'Lightweight and powerful wsgi rest framework for rapid building applications based on wsgi servers.',
    'long_description': '# Chocs [![PyPI version](https://badge.fury.io/py/chocs.svg)](https://pypi.org/project/chocs/) ![Release](https://github.com/kodemore/chocs/workflows/Release/badge.svg) ![Linting and Tests](https://github.com/kodemore/chocs/workflows/Linting%20and%20Tests/badge.svg) [![codecov](https://codecov.io/gh/kodemore/chocs/branch/master/graph/badge.svg)](https://codecov.io/gh/kodemore/chocs) [![Maintainability](https://api.codeclimate.com/v1/badges/9e3c979283b2361a9174/maintainability)](https://codeclimate.com/github/kodemore/chocs/maintainability)\nChocs is a modern HTTP framework for building AWS HTTP API/REST API and WSGI compatible applications. \nChocs aims to be small, expressive, and robust. \nIt provides an elegant API for writing fault-proof, extensible microservices.  \n\n## Features\n\n - AWS Serverless integration\n - Open api integration  \n - Elegant and easy API\n - No additional bloat like built-in template engines, session handlers, etc.\n - Compatible with all WSGI servers\n - Loosely coupled components which can be used separately\n - Multipart body parsing\n - Graceful error handling\n - HTTP middleware support\n - Fast routing\n\n## Installation\n```\npip install chocs\n```\n\n# Usage\n\n## Quick start\n\n```python\nfrom chocs import Application\nfrom chocs import HttpRequest\nfrom chocs import HttpResponse\nfrom chocs import serve\n\n\nhttp = Application()\n\n@http.get("/hello/{name}")\ndef hello(request: HttpRequest) -> HttpResponse:\n    return HttpResponse(f"Hello {request.path_parameters.get(\'name\')}!")\n\nserve(http)\n```\n\n > Keep in mind that the `serve()` function is using the `bjoern` package, so make sure you included it in your project \n > dependencies before using it. You are able to use any WSGI compatible server.\n\n## Running application with Gunicorn (or any other WSGI server)\n\n```python\n# myapp.py\nfrom chocs import Application\nfrom chocs import HttpRequest\nfrom chocs import HttpResponse\nfrom chocs import create_wsgi_handler\n\n\nhttp = Application()\n\n\n@http.get("/hello/{name}*")\ndef hello(request: HttpRequest) -> HttpResponse:\n    return HttpResponse(f"Hello {request.path_parameters.get(\'name\')}!")\n\napp = create_wsgi_handler(http, debug=False)\n```\n\n```bash\ngunicorn -w 4 myapp:app\n```\n\n## Running application in AWS Lambda (Http api or rest api)\n\n`handler.py`\n```python\nimport logging\n\nfrom chocs import HttpRequest\nfrom chocs import HttpResponse\nfrom chocs import Application\n\nlogger = logging.getLogger()\nlogger.setLevel(logging.INFO)\n\n\nhttp = Application()\n\n\n@http.get("/hello/{name}")\ndef hello_handler(request: HttpRequest) -> HttpResponse:\n    logger.info("Hello AWS!")\n    logger.info(request.attributes.get("aws_context"))\n    logger.info(request.attributes.get("aws_event"))\n    return HttpResponse(f"Hello {request.path_parameters.get(\'name\')}")\n\n\n__all__ = ["hello_handler"]\n```\n\n`serverless.yml`\n```yaml\nservice: aws-hello-name\n\nprovider:\n  name: aws\n  runtime: python3.8\n\nplugins:\n  - serverless-python-requirements\n\ncustom:\n  pythonRequirements:\n    dockerizePip: true\n\nfunctions:\n  hello_name:\n    handler: handler.hello_handler\n    events:\n      - httpApi:\n          method: GET\n          path: /hello/{name}\n```\n\n```bash\nserverless deploy\n```\n\n## Routing\nChocs is shipped with a built-in routing module. The easiest way to utilise chocs\' routing is to use `chocs.router` object.\n`chocs.router` is an instance of the module\'s internal class `chocs.application.Application`, which provides a simple API \nwhere each function is a decorator corresponding to an HTTP method.\n\n```python\nfrom chocs import Application, HttpResponse, HttpRequest\n\n\nhttp = Application()\n\n\n@http.get("/hello")\ndef hello(req: HttpRequest) -> HttpResponse:\n    ...\n```\n\nThe above example will assign the hello function to handle a `GET /hello` request. \n\nAvailable methods:\n- `delete`\n- `get`\n- `head`\n- `options`\n- `patch`\n- `post`\n- `put`\n- `trace`\n\n### Parametrized routes\n\nRoutes can contain parameterised parts. Parameters must be enclosed within `{` and `}`.\n\n```python\nfrom chocs import Application\n\nhttp = Application()\n\n\n@http.get("/pet/{id}")\ndef hello():\n    ...\n```\nWill match the following URIs:\n - `/pet/1`\n - `/pet/abc`\n - `/pet/abc1`\n \n### Wildcard routes\n\nAsterisks (`*`) can be used in the route\'s pattern to match any possible combination. Keep in mind that routes which \n_do not_ contain wildcards are prioritised over routes with wildcards.\n\n```python\nfrom chocs import Application\n\nhttp = Application()\n\n\n@http.get("/pet/*", id)\ndef hello():\n    ...\n```\n\nThe above example will match following URIs:\n- `/pet/a`\n- `/pet/a/b/c`\n- `/pet/12jd/fds`\n\n### Route groups\n\nChocs supports route groups. Route groups is implemented through [context lib interface](https://docs.python.org/3/library/contextlib.html).\nIf you need to split your application in smaller chunks with standalone req/res handlers consider the\nfollowing example:\n\n```python\nfrom threading import Thread\n\nfrom chocs.wsgi import serve \nfrom chocs import Application\nfrom chocs import HttpRequest\nfrom chocs import HttpResponse\n\nmain_app = Application()\n\nwith main_app.group("/users/{id}") as user_module:\n    \n    @user_module.post("/profile_picture")  # POST /users/{id}/profile_pictures\n    def create_profile_picture(request: HttpRequest) -> HttpResponse:\n        ...\n    \n    @user_module.get("/profile_picture")  # GET /users/{id}/profile_pictures\n    def get_profile_picture(request: HttpRequest) -> HttpResponse:\n        ...\n    \n    @user_module.get("/badges") # GET /users/{id}/badges\n    def badges(request: HttpRequest) -> HttpResponse:\n        ...\n\nwith main_app.group("/payments") as payment_module:\n\n    @payment_module.get("/analytics") # GET /payments/analytics\n    def get_analytics(request: HttpRequest) -> HttpResponse:\n        ...\n\nif __name__ == \'__main__\':\n    def wsgi_user_module():\n        serve(user_module, port=8081)\n    def wsgi_payment_module():\n        serve(payment_module, port=8082)\n\n    Thread(target=wsgi_user_module).start()\n    payment_module()\n```\n\nThe above example shows how to run two different modules, which support their own routes\non two different ports in the one process.\n\n## Middleware\n\nMiddleware are functions or classes that inherit `chocs.Middleware`. Middlewares have access to the request object\nand the `next` function which can be used to control middleware stack flow. Successful middleware execution should call\nthe `next` function which accepts a `chocs.HttpRequest` instance and returns `chocs.HttpReponse`.\n\nMiddlewares can perform various tasks:\n- Making changes in request/response objects ending\n- Validating input data\n- Authenticating users\n- End request-response cycle\n- Connecting to external data sources\n\n## Integration with openapi\n\nChocs provides middleware which can be used to validate input data and simplify working with \ninputs by mapping them to dataclasses. To provide automatic validation for your request based\non open api specification, use `chocs.middleware.OpenApiMiddleware`:\n\n```python\nfrom chocs.middleware import OpenApiMiddleware\nfrom chocs import Application, HttpRequest, HttpResponse\nfrom os import path\nfrom dataclasses import dataclass\n\n# absolute path to file containing open api documentation; yaml and json files are supported\nopenapi_filename = path.join(path.dirname(__file__), "/openapi.yml")\n\n# instantiating application and passing open api middleware\napp = Application(OpenApiMiddleware(openapi_filename, validate_body=True, validate_query=True))\n\n# defining our dataclass for better type support\n@dataclass()\nclass Pet:\n    id: str\n    name: str\n\n# the registered route must correspond to open api route within `path` section.\n# if request body is invalid the registered function will not be executed\n@app.post("/pets", parsed_body=Pet, strict=True) # `parsed_body` parameter can be used to map request to certain type\ndef create_pet(request: HttpRequest) -> HttpResponse:\n    try: # mapping request to your dataclass may fail\n        pet: Pet = request.parsed_body\n    except TypeError:\n        return HttpResponse(status=400)\n    \n    # if request body is valid `request.parsed_body` will be instance of a Pet\n    assert isinstance(pet, Pet)\n    return HttpResponse(pet.name)\n```\n\n> Open api file used in the example above can be [found here](./examples/input_validation_with_open_api/openapi.yml)\n\nChocs automatically validates:\n - request body, `application/json` header must be present for successful validation\n - query string parameters\n - request headers\n\n### Turning strict mode off\n\n> Note: By default chocs works in a strict mode, which means when you map request\n> data to your object `__init__` method is called. To override this behaviour set\n> `strict` property to false: `@app.post("/pets", parsed_body=Pet, strict=False)`\n\nIn non strict mode your request body will be still mapped to the provided class, and data will be hydrated but `__init__` method will not be called.\nPlease consider the following example:\n\n```python\nfrom chocs.middleware import OpenApiMiddleware\nfrom chocs import Application, HttpRequest, HttpResponse\nfrom os import path\nfrom dataclasses import dataclass\n\nopenapi_filename = path.join(path.dirname(__file__), "/openapi.yml")\napp = Application(OpenApiMiddleware(openapi_filename, validate_body=True, validate_query=True))\n\n@dataclass()\nclass Pet:\n    id: str\n    name: str\n\n@app.post("/pets", parsed_body=Pet, strict=False) # with strict mode off, __init__ method of Pet\'s class is ignored.\ndef create_pet(request: HttpRequest) -> HttpResponse:\n    pet: Pet = request.parsed_body # this will always work\n\n    assert isinstance(pet, Pet)\n    return HttpResponse(pet.name)\n```\n\n\n### Handling validation errors with custom middleware\n\nBy default, if validation fails users will see `500 response`. This behavior can be changed if custom middleware that\ncatches validation errors is defined and used in application.\n\n### Defining and using a custom middleware\n \nThe following code defines simple function middleware to catch validation errors when they appear and notifies users:\n\n```python\nfrom chocs.middleware import OpenApiMiddleware\nfrom chocs.json_schema.errors import ValidationError\nfrom chocs import Application, HttpRequest, HttpResponse\nfrom dataclasses import dataclass\nimport json\nfrom typing import Callable\nfrom os import path\n\nopenapi_filename = path.join(path.dirname(__file__), "/openapi.yml")\n\n\n# middleware must always accept two parameters; HttpRequest and Callable and return HttpResponse\ndef handle_errors(request: HttpRequest, next: Callable) -> HttpResponse:\n    try:\n        return next(request) # we pass request further to middleware pipeline\n    except ValidationError as error: # if exception is thrown it is caught here and new response is generated instead\n        json_response = {\n            "code": error.code,\n            "message": str(error),\n        }\n        return HttpResponse(json.dumps(json_response), status=422)\n    \n# error handling middleware must go before open api one to catch errors thrown inside open api middleware\napp = Application(handle_errors, OpenApiMiddleware(openapi_filename))\n\n@dataclass()\nclass Pet:\n  id: str\n  name: str\n\n@app.post("/pets", parsed_body=Pet)\ndef create_pet(request: HttpRequest) -> HttpResponse:\n  assert isinstance(request.parsed_body, Pet)\n\n  pet = request.parsed_body\n  return HttpResponse(pet.name)\n```\n\nFull working example can be found inside [examples directory](./examples/input_validation_with_open_api)\n\n## Request\n`chocs.Request` object is an abstraction around WSGI\'s environment and `wsgi.input` data with handy interface \nto ease everyday work.\n\n#### `chocs.Request.headers:chocs.HttpHeaders (read-only)`\nKeeps parsed headers in dict-like object.\n\n#### `chocs.Request.body:io.BytesIO` \nRaw body data\n\n#### `chocs.Request.parsed_body:chocs.HttpMessage`\nDepending on the content type it could be one of the following:\n - `chocs.FormHttpMessage`\n - `chocs.JsonHttpMessage`\n - `chocs.MultipartHttpMessage`\n - `chocs.YamlHttpMessage`\n\n#### `chocs.Request.as_dict(): dict`\nTries to convert request body to a dict and returns it.\n\n> Note this will only work with json and yaml content types.\n\n#### `chocs.Request.as_str(): str`\nReturns request content as a string.\n \n#### `chocs.Request.cookies:typing.List[chocs.HttpCookie]` \nRequest\'s cookies\n\n#### `chocs.Request.method:chocs.HttpMethod`\nThe request\'s method\n\n#### `chocs.Request.path:str`\nThe request\'s path\n\n#### `chocs.Request.query_string:chocs.HttpQueryString`\nA dict like object with parsed query string with JSON forms support\n        \n#### `chocs.Request.path_parameters:dict`\nMatched route parameters, for example when `/users/john` matches the `/users/{name}` route, parameters will contain a \n`name` key with a value of `john`\n\n#### `chocs.Request.attributes:dict`\nOther environmental or custom attributes attached to the request object, eg.: `aws_event` or `aws_context`\nwhen running chocs app as aws lambda.\n\n## Response\n`chocs.Response` object is a part of request-response flow and it is required to be returned by all functions\ndecorated with `router.*` method. Instance of the response class is recognised by `chocs.Application` and used to \ngenerate real response served to your clients.\n\n#### `chocs.Response.body: io.BytesIO` \nBody served to server\'s clients.\n\n### `chocs.Response.status_code: chocs.HttpStatus`\nValid response code, instance of `chocs.HttpStatus` enum can be used or just a status code\'s number.\n\n#### `chocs.Response.cookies:chocs.HttpCookieJar` \nResponse\'s cookies\n\n#### `chocs.Response.write(body: Union[bytes, str, bytearray])`\nWrite bytes to response body\n\n#### `chocs.Response.close()`\nMakes body non-writable.\n\n#### `chocs.Response.writable: bool`\nIndicates whether response\'s body is writable.\n\n#### `chocs.Response.parsed_body:chocs.HttpMessage`\nDepending on the content type it could be one of the following:\n- `chocs.FormHttpMessage`\n- `chocs.JsonHttpMessage`\n- `chocs.MultipartHttpMessage`\n- `chocs.YamlHttpMessage`\n\n#### `chocs.Response.as_dict(): dict`\nTries to convert response body to a dict and returns it.\n\n> Note this will only work with json and yaml content types.\n\n#### `chocs.Response.as_str(): str`\nReturns response content as a string.\n\n## Working with cookies\n\n`chocs.HttpCookieJar` object takes care of cookie handling. It can be accessed in dict-like manner, when item is requested,\ninstance of `chocs.HttpCookie` is returned to user. \n\nCookies can be set either by passing string value to the `chocs.CookieJar`\'s key, or by calling `chocs.CookieJar.append` \nmethod which accepts instance of `chocs.Cookie`.\n\n### Reading client cookies\n\nCookies can be easily accessed from `chocs.Request.cookies` object which is injected as a parameter to each function \nregistered as route handler. Consider the following example:\n\n```python\nfrom chocs import HttpRequest\nfrom chocs import HttpResponse\nfrom chocs import Application\nfrom chocs import serve\n\nhttp = Application()\n\n\n@http.get("/cookies")\ndef read_cookies(request: HttpRequest) -> HttpResponse:\n\n    message = "Hello"\n    if "user_name" in request.cookies:\n        message += f", {str(request.cookies[\'user_name\'])}"\n    message += "!"\n\n    return HttpResponse(body=message)\n\nserve(http)\n```\n\n### Setting cookies\n```python\nfrom datetime import datetime\n\nfrom chocs import HttpCookie\nfrom chocs import HttpRequest\nfrom chocs import HttpResponse\nfrom chocs import Application\nfrom chocs import serve\n\nhttp = Application()\n\n\n@http.get("/cookies")\ndef read_cookies(request: HttpRequest) -> HttpResponse:\n    response = HttpResponse(body="Hi! I have baked some cookies for ya!")\n    response.cookies[\'simple-cookie\'] = "Simple cookie for simple people"\n    response.cookies.append(HttpCookie("advanced-cookie", "This cookie will expire in 2021-01-01", expires=datetime(2021, 1, 1)))\n    return response\n\nserve(http)\n```\n\n# Contributing\n\n## Prerequisites\n\n- libev\n- python 3.8\n- docker\n\n## Installation\n\n`poetry install`\n\n## Running tests\n\n`poetry run pytest`\n\n## Linting\n\n```shell\npoetry run black .\npoetry run isort .\npoetry run mypy .\n```\n\n## PR \n',
    'author': 'Dawid Kraczkowski',
    'author_email': 'dawid.kraczkowski@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kodemore/chocs',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Callable

from chocs.http_request import HttpRequest
from chocs.http_response import HttpResponse
from chocs.middleware.middleware import MiddlewarePipeline
from chocs.routing import Route

from .aws import AwsServerlessFunction
from .serverless import IS_AWS_ENVIRONMENT, ServerlessFunction


def create_serverless_function(
    func: Callable[[HttpRequest], HttpResponse], route: Route, middleware_pipeline: MiddlewarePipeline,
) -> Callable:

    if IS_AWS_ENVIRONMENT:
        return AwsServerlessFunction(func, route, middleware_pipeline)

    return ServerlessFunction(func, route, middleware_pipeline)


__all__ = ["create_serverless_function"]
